unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, MMSystem, WebcamCapture, BallDetect, RCXBasics,
  RCXFunctions;

type
  TDirection = (diForwards, diBackwards, diRight, diLeft);

  TfrmMain = class(TForm)
    pnlContainer: TPanel;
    lblCopyright: TLabel;
    pnlWebcam: TPanel;
    gbVideo: TGroupBox;
    imgCapture: TImage;
    cbShowVideo: TCheckBox;
    cbSearchForBall: TCheckBox;
    btnVideoSource: TButton;
    gbRobotControl: TGroupBox;
    lblBytecodeReplyText: TLabel;
    lblBytecodeReply: TLabel;
    btnStopSearching: TButton;
    btnForward: TBitBtn;
    btnRight: TBitBtn;
    btnLeft: TBitBtn;
    btnBackward: TBitBtn;
    btnEnableJoystick: TCheckBox;
    edBytecode: TEdit;
    btnSendBytecode: TButton;
    btnStartSearching: TButton;
    gbStatus: TGroupBox;
    lblBallDistanceText: TLabel;
    lblBallSizeText: TLabel;
    lblBallDistance: TLabel;
    lblBallSize: TLabel;
    gbCalibrateCam: TGroupBox;
    lblCamHeight: TLabel;
    lblMinDistance: TLabel;
    lblMaxHeight: TLabel;
    lblCm1: TLabel;
    lblCm2: TLabel;
    lblCm3: TLabel;
    edCamHeight: TEdit;
    edMinDistance: TEdit;
    edMaxHeight: TEdit;
    timCapture: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure timCaptureTimer(Sender: TObject);
    procedure btnStartSearchingClick(Sender: TObject);
    procedure btnVideoSourceClick(Sender: TObject);
    procedure cbShowVideoClick(Sender: TObject);
    procedure btnForwardMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnForwardMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnLeftMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnRightMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnBackwardMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnSendBytecodeClick(Sender: TObject);
    procedure btnStopSearchingClick(Sender: TObject);
    procedure edCamHeightChange(Sender: TObject);
    procedure edBytecodeKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure btnEnableJoystickClick(Sender: TObject);
  private
    { Private declarations }
    Webcam: TWebcamCapture;
    FProcessing: Boolean;   // TRUE: program is processing a captured image
    FRobotActive: Boolean;  // TRUE: robot is actively searching for red balls
    FCamHeight: Single;    // height of the camera lens (in centimeters)
    FMinDistance: Single;  // minimum distance that can be captured by the camera (in cm)
    FMaxHeight: Single;    // maximum height that can be captured at the minimum distance (in cm)
    function CalculateDistance(YPixel: Integer): Integer;
    function CalculateDiameter(YPixelBottom, YPixelTop: Integer): Integer;
    procedure DriveTo(BallData: TBallData);
    procedure OnCaptureBitmap(Bitmap: TBitmap);
    procedure RunRobot(Direction: TDirection; Duration: Integer);
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.DFM}
{$R sound.res}

procedure CheckTowerStatus;
begin
  if OpenUSBTowerComm = False then
    MessageBox(frmMain.Handle, 'Verbindung zu USB-Tower konnte nicht hergestellt werden!' , PChar(LowerCase(ExtractFileName(Application.exename))), MB_APPLMODAL or MB_OK or MB_ICONEXCLAMATION);
end;

// CalculateDistance
// Returns the distance of the ball relative to the webcam
// YPixel: vertical position within the image (in pixels)
function TfrmMain.CalculateDistance(YPixel: Integer): Integer;
var
  YImg: Single;
begin
  YImg := (FMaxHeight / 240) * (240 - YPixel);
  Result := Trunc((FMinDistance * FCamHeight) / (FCamHeight - YImg));
end;

// CalculateDiameter
// Returns the diameter of the ball
function TfrmMain.CalculateDiameter(YPixelBottom, YPixelTop: Integer): Integer;
var
  YImgBottom, YImgTop: Single;
begin
  YImgBottom := (FMaxHeight / 240) * (240 - YPixelBottom);
  YImgTop := (FMaxHeight / 240) * (240 - YPixelTop);
  Result := Trunc(FCamHeight + (((YImgTop - FCamHeight) * FCamHeight) / (FCamHeight - YImgBottom)));
end;

procedure TfrmMain.DriveTo(BallData: TBallData);
var
  Distance: Integer;
begin
  Distance := CalculateDistance(BallData.YPos + BallData.Radius);
  if Distance <= FMinDistance + 2 then
  begin
    frmMain.btnStopSearchingClick(nil);
    PlaySound('alert', hInstance, SND_ASYNC or SND_MEMORY or SND_RESOURCE);
    exit;
  end;

  if BallData.XPos < 140 then
  begin
    RunRobot(diLeft, 0);
  end else
  if BallData.XPos > 180 then
  begin
    RunRobot(diRight, 0);
  end else
  begin
    RunRobot(diForwards, Round(((Distance - FMinDistance) / 3) * 100));
  end;
end;

procedure TfrmMain.OnCaptureBitmap(Bitmap: TBitmap);
var
  BallData: TBallData;
  Time: Integer;
begin
  FProcessing := True;

  if cbSearchForBall.Checked = True then
  begin
    BallData := FindRedBall(Bitmap);

    // if a ball (or rather a filled circle) was found
    if (BallData.XPos > 0) and (BallData.YPos > 0) and (BallData.Radius > 0) then
    begin
      Bitmap.Canvas.Brush.Style := bsClear;
      Bitmap.Canvas.Pen.Color := clBlue;
      Bitmap.Canvas.Pen.Width := 5;
      Bitmap.Canvas.Ellipse(BallData.XPos - BallData.Radius,
        BallData.YPos - BallData.Radius,
        BallData.XPos + BallData.Radius,
        BallData.YPos + BallData.Radius);
      lblBallDistance.Caption := IntToStr(CalculateDistance(BallData.YPos + BallData.Radius)) + ' cm';
      lblBallSize.Caption := IntToStr(CalculateDiameter(BallData.YPos + BallData.Radius, BallData.YPos - BallData.Radius)) + ' cm';
      imgCapture.Canvas.Draw(0, 0, Bitmap);
      imgCapture.Canvas.Refresh;
      Application.ProcessMessages;
      if FRobotActive = True then
        DriveTo(BallData);
    end else
    begin
      imgCapture.Canvas.Draw(0, 0, Bitmap);
      imgCapture.Canvas.Refresh;
      Application.ProcessMessages;
      if FRobotActive = True then
        RunRobot(diLeft, 0);
    end;
  end else
    imgCapture.Canvas.Draw(0, 0, Bitmap);

  if FRobotActive = True then
  begin
    // wait for 1 second to allow the camera to focus accurately
    Time := GetTickCount;
    while GetTickCount - Time < 1000 do
    begin
      Application.ProcessMessages;
    end;
  end;

  FProcessing := False;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  imgCapture.Picture.Bitmap.PixelFormat := pfDevice;

  DecimalSeparator := ',';
  try
    FCamHeight := StrToFloat(edCamHeight.Text);
    FMinDistance := StrToFloat(edMinDistance.Text);
    FMaxHeight := StrToFloat(edMaxHeight.Text);
  except
  end;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  if Assigned(Webcam) then
    Webcam.Free;      // free Webcam object
  CloseUSBTowerComm;  // close connection to USB Tower
end;

procedure TfrmMain.timCaptureTimer(Sender: TObject);
begin
  if FProcessing = False then
    Webcam.CaptureBitmap;
end;

procedure TfrmMain.RunRobot(Direction: TDirection; Duration: Integer);
var
  Time: Integer;
begin
  if Duration < 0 then
    Duration := 0;

  case Direction of
    diForwards:
      begin
        SetMotorDirection(RCXMotorA, RCXForwards);
        SetMotorDirection(RCXMotorC, RCXBackwards);
      end;
    diBackwards:
      begin
        SetMotorDirection(RCXMotorA, RCXBackwards);
        SetMotorDirection(RCXMotorC, RCXForwards);
      end;
    diLeft:
      begin
        SetMotorDirection(RCXMotorA, RCXForwards);
        SetMotorDirection(RCXMotorC, RCXForwards);
      end;
    diRight:
      begin
        SetMotorDirection(RCXMotorA, RCXBackwards);
        SetMotorDirection(RCXMotorC, RCXBackwards);
      end;
  end;

  if StartMotors([RCXMotorA, RCXMotorC]) = False then
    StartMotors([RCXMotorA, RCXMotorC]);

  Time := GetTickCount;
  while GetTickCount - Time < Duration do
  begin
    Application.ProcessMessages;
  end;
  if StopMotors([RCXMotorA, RCXMotorC]) = False then
    StopMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnStartSearchingClick(Sender: TObject);
begin
  CheckTowerStatus;

  btnStopSearching.Visible := True;
  btnStartSearching.Visible := False;

  FRobotActive := True;

  cbShowVideo.Enabled := False;
  cbSearchForBall.Enabled := False;
  cbSearchForBall.Checked := True;
  if cbShowVideo.Checked = False then
    cbShowVideo.Checked := True;
end;

procedure TfrmMain.btnVideoSourceClick(Sender: TObject);
begin
  Webcam.DlgVideoSource;
end;

procedure TfrmMain.cbShowVideoClick(Sender: TObject);
begin
  if (cbShowVideo.Checked = True) and (Assigned(Webcam) = False) then
  begin
    Screen.Cursor := crHourglass;
    Application.ProcessMessages;
    Webcam := TWebcamCapture.Create(320, 240, pnlWebcam.Handle);
    Webcam.OnCaptureBitmap := OnCaptureBitmap;
    Screen.Cursor := crDefault;
  end;

  timCapture.Enabled := cbShowVideo.Checked;
  btnVideoSource.Enabled := cbShowVideo.Checked;
  cbSearchForBall.Enabled := (cbShowVideo.Checked) and (btnStartSearching.Visible = True);
end;

procedure TfrmMain.btnForwardMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  CheckTowerStatus;
  SetMotorDirection(RCXMotorA, RCXForwards);
  SetMotorDirection(RCXMotorC, RCXBackwards);
  StartMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnForwardMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  StopMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnLeftMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CheckTowerStatus;
  SetMotorDirection(RCXMotorA, RCXForwards);
  SetMotorDirection(RCXMotorC, RCXForwards);
  StartMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnRightMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CheckTowerStatus;
  SetMotorDirection(RCXMotorA, RCXBackwards);
  SetMotorDirection(RCXMotorC, RCXBackwards);
  StartMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnBackwardMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  CheckTowerStatus;
  SetMotorDirection(RCXMotorA, RCXBackwards);
  SetMotorDirection(RCXMotorC, RCXForwards);
  StartMotors([RCXMotorA, RCXMotorC]);
end;

procedure TfrmMain.btnSendBytecodeClick(Sender: TObject);
var
  Bytes: String;
begin
  CheckTowerStatus;

  Bytes := FormatHexToString(Trim(edBytecode.Text));
  if Bytes = '' then
  begin
    MessageBox(frmMain.Handle, 'Ungltige Eingabe!' , PChar(LowerCase(ExtractFileName(Application.exename))), MB_APPLMODAL or MB_OK or MB_ICONEXCLAMATION);
  end else
  begin
    lblBytecodeReply.caption := FormatStringToHex(SendPacket(CreatePacket(Bytes)));
  end;
end;



procedure TfrmMain.btnStopSearchingClick(Sender: TObject);
begin
  btnStartSearching.Visible := True;
  btnStopSearching.Visible := False;

  FRobotActive := False;

  cbShowVideo.Checked := False;
  cbShowVideo.Enabled := True;
end;

procedure TfrmMain.edCamHeightChange(Sender: TObject);
begin
  try
    FCamHeight := StrToFloat(edCamHeight.Text);
    FMinDistance := StrToFloat(edMinDistance.Text);
    FMaxHeight := StrToFloat(edMaxHeight.Text);
  except
    on E: EConvertError do
    begin
      MessageBox(frmMain.Handle, 'Ungltiger Wert!' , PChar(LowerCase(ExtractFileName(Application.exename))), MB_APPLMODAL or MB_OK or MB_ICONEXCLAMATION);
      exit;
    end;
  end;
end;

procedure TfrmMain.edBytecodeKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then
    btnSendBytecodeClick(Sender);
end;

procedure TfrmMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_F5 then
    // switch to full screen mode
    if frmMain.BorderStyle = bsSingle then
    begin
      frmMain.Visible := False;
      frmMain.BorderStyle := bsNone;
      frmMain.Left := 0;
      frmMain.Top := 0;
      frmMain.Width := Screen.Width;
      frmMain.Height := Screen.Height;
      frmMain.Color := clBlack;
      pnlContainer.Left := (frmMain.Width - pnlContainer.Width) DIV 2;
      pnlContainer.Top := (frmMain.Height - pnlContainer.Height) DIV 2;
      frmMain.Visible := True;
    end else
    begin
      frmMain.Visible := False;
      frmMain.BorderStyle := bsSingle;
      frmMain.Width := 663;
      frmMain.Height := 455;
      frmMain.Left := (Screen.Width - frmMain.Width) DIV 2;
      frmMain.Top := (Screen.Height - frmMain.Height) DIV 2;
      frmMain.Color := clBtnFace;
      pnlContainer.Left := 0;
      pnlContainer.Top := 0;
      frmMain.Visible := True;
    end;
end;

procedure TfrmMain.btnEnableJoystickClick(Sender: TObject);
begin
  // not yet implemented
end;

end.
