{*******************************************************}
{                                                       }
{  RCXBasics                                            }
{  Copyright (c) 2003 by Andreas Wulf                   }
{                                                       }
{*******************************************************}
{                                                       }
{  Basic functions for communicating with an RCX        }
{  (a programmable, microcontroller-based brick) via    }
{  the LEGO USB Tower.                                  }
{                                                       }
{  Last modified: Oct 19, 2003                          }
{                                                       }
{*******************************************************}

unit RCXBasics;

interface

uses
  Windows, SysUtils;

var
  TowerHandle: THandle;
  LastCommand: Char;

function OpenUSBTowerComm: Boolean;
procedure CloseUSBTowerComm;
function CreatePacket(Msg: String): String;
function SendPacket(Packet: String): String;
function FormatStringToHex(Msg: String): String;
function FormatHexToString(Msg: String): String;

implementation

function OpenUSBTowerComm: Boolean;
begin
  if TowerHandle <= 0 then
    TowerHandle := CreateFile('\\.\LEGOTOWER1', GENERIC_WRITE or GENERIC_READ, FILE_SHARE_WRITE or FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);

  Result := TowerHandle <> INVALID_HANDLE_VALUE;
end;

procedure CloseUSBTowerComm;
begin
  if TowerHandle > 0 then
  begin
    CloseHandle(TowerHandle);
    TowerHandle := 0;
  end;
end;

function IsValidPacket(Packet: String): Boolean;
var
  I, Checksum: Integer;
begin
  Result := False;

  // RCX packet structure: 0x55 0xFF 0x00 D1 ~D1 D2 ~D2 ... Dn ~Dn C ~C
  if (Length(Packet) < 7)
    or ((Length(Packet) - 3) mod 2 = 1)
    or (Copy(Packet, 1, 3) <> #$55 + #$FF + #$00)
    or (Char(not Ord(Packet[4])) <> LastCommand) then
    exit;

  // check if inversions are valid and calculate packet checksum
  Checksum := 0;
  I := 4;
  while I < length(Packet) do
  begin
    if Packet[I] <> Char(not Ord(Packet[I + 1])) then
      exit;
    if I < length(Packet) - 1 then
      Checksum := Checksum + Ord(Packet[I]);
    I := I + 2;
  end;

  // check if checksum is valid
  if (Checksum and $FF) <> Ord(Packet[length(Packet) - 1]) then
    exit;

  Result := True;
end;

function CreatePacket(Msg: String): String;
var
  I, Checksum: Integer;
begin
  if Msg = '' then
  begin
    Result := '';
    exit;
  end;

  if Msg[1] = LastCommand then
    Msg[1] := Char(Ord(Msg[1]) xor 8);
  LastCommand := Msg[1];

  Result := #$55 + #$FF + #$00;
  Checksum := 0;
  for I := 1 to length(Msg) do
  begin
    Result := Result + Msg[I] + Char((not Ord(Msg[I])));
    Checksum := Checksum + Ord(Msg[I]);
  end;
  Result := Result + Char(Checksum) + Char(not Checksum);
end;

function ReceivePacket: String;
var
  Buffer: array [0..255] of Char;
  NumRead, I: Integer;
begin
  Result := '';
  if ReadFile(TowerHandle, Buffer, SizeOf(Buffer), NumRead, nil) = True then
    for I := 0 to NumRead - 1 do
      Result := Result + Buffer[I];

  // if IsValidPacket(Result) = False then
  //   Result := '';
end;

function SendPacket(Packet: String): String;
var
  I, NumWritten: Integer;
begin
  Result := '';

  // try 3 times to send the packet and to receive an answer
  for I := 1 to 3 do
  begin
    if WriteFile(TowerHandle, PChar(Packet)^, Length(Packet), NumWritten, nil) = False then
      Continue;
    if NumWritten <> Length(Packet) then
      Continue;
    Result := ReceivePacket;
    if Result <> '' then
      Break;
  end;
end;

function FormatStringToHex(Msg: String): String;
var
  I: Integer;
begin
  Result := '';
  for I := 1 to length(Msg) do
  begin
    Result := Result + '0x' + IntToHex(Ord(Msg[I]), 2);
    if I < length(Msg) then
      Result := Result + ' ';
  end;
end;

function FormatHexToString(Msg: String): String;
begin
  Result := '';
  Msg := Msg + ' ';
  while (Copy(Msg, 1, 2) = '0x') and (Copy(Msg, 5, 1) = ' ') do
  begin
    try
      Result := Result + Char(StrToInt('$' + Copy(Msg, 3, 2)));
    except
      Result := '';
      exit;
    end;
    Delete(Msg, 1, 5);
  end;
  if Msg <> '' then
    Result := '';
end;

end.
